import java.io.IOException;
import java.util.Iterator;

import org.apache.bcel.Constants;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.BasicType;
import org.apache.bcel.generic.BranchHandle;
import org.apache.bcel.generic.CodeExceptionGen;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.InstructionFactory;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.Type;

/**
 * This class performs transformation of the given class. Added code counts
 * executed instructions.
 *
 * @author Kuba Madej
 *
 */
public class Transform {
	/**
	 * ShutdownHook's method that increment invocations count
	 */
	private static final String INCREMENT_INVOCATIONS_COUNT_METHOD = "incrementInvocationsCount";

	/**
	 * Return code of the preoces in case of failure
	 */
	private static final int FAILURE_CODE = 1;

	/**
	 * Class file extension
	 */
	private static final String CLASS_FILE_EXTENSION = ".class";

	/**
	 * Class to be transformed
	 */
	private JavaClass javaClass;

	/**
	 * Constant pool generator of class to be transformed
	 */
	private final ConstantPoolGen constantPoolGen;

	/**
	 * InstructionFactory to be used
	 */
	private final InstructionFactory instructionFactory;

	/**
	 * Constructs new transformer
	 *
	 * @param className
	 *            name of the class to be transformed
	 */
	public Transform(final String className) {
		try {
			javaClass = Repository.lookupClass(className);
		} catch (ClassNotFoundException e) {
			System.err.println("Class not found");
			System.exit(FAILURE_CODE);
		}

		constantPoolGen = new ConstantPoolGen(javaClass.getConstantPool());
		instructionFactory = new InstructionFactory(constantPoolGen);
	}

	/**
	 * Starts the transofmer
	 *
	 * @param args
	 *            arguments array
	 */
	public static void main(final String[] args) {
		if (args.length < 1) {
			System.err.println("Usage: java Transform <classname>");
			System.exit(0);
		}
		new Transform(args[0]).transform();
	}

	/**
	 * Performs the actual transformation
	 *
	 */
	public void transform() {
		final Method[] methods = javaClass.getMethods();

		for (int i = 0; i < methods.length; i++) {
			final Method method = methods[i];
			if (!(method.isNative() || method.isAbstract())) {
				final String methodName = method.getName();
				if (!(methodName.equals(Constants.CONSTRUCTOR_NAME) || methodName
						.equals(Constants.STATIC_INITIALIZER_NAME)))
					methods[i] = transform(new MethodGen(method, javaClass
							.getClassName(), constantPoolGen));
			}

		}

		javaClass.setConstantPool(constantPoolGen.getFinalConstantPool());

		try {
			javaClass.dump(javaClass.getClassName() + CLASS_FILE_EXTENSION);
		} catch (IOException e) {
			System.err.println("Couldn't write class file");
			System.exit(FAILURE_CODE);
		}
	}

	/**
	 * Transforms a single method
	 *
	 * @param methodGen
	 *            MethodGen instance of the method to be tranformed
	 * @return transformed method
	 */
	private Method transform(final MethodGen methodGen) {
		final InstructionList instructionList = methodGen.getInstructionList();

		for (Iterator iter = instructionList.iterator(); iter.hasNext();) {
			InstructionHandle instructionHandle = (InstructionHandle) iter
					.next();

			InstructionList instructionsToInsert = new InstructionList();
			instructionsToInsert.append(instructionFactory
					.createConstant(instructionHandle.getInstruction()
							.getName()));
			instructionsToInsert.append(instructionFactory.createInvoke(
					ShutdownHook.class.getName(),
					INCREMENT_INVOCATIONS_COUNT_METHOD, BasicType.VOID,
					new Type[] { ObjectType.STRING }, Constants.INVOKESTATIC));

			instructionList.insert(instructionHandle, instructionsToInsert);
		}

		for (Iterator iter = instructionList.iterator(); iter.hasNext();) {
			final InstructionHandle instructionHandle = (InstructionHandle) iter
					.next();
			if (instructionHandle instanceof BranchHandle) {
				BranchHandle branchHandle = (BranchHandle) instructionHandle;
				branchHandle.setTarget(branchHandle.getTarget().getPrev()
						.getPrev());
			}
		}

		for (CodeExceptionGen codeExceptionGen : methodGen
				.getExceptionHandlers()) {
			codeExceptionGen.setHandlerPC(codeExceptionGen.getHandlerPC()
					.getPrev().getPrev());
		}

		methodGen.setMaxStack();
		return methodGen.getMethod();
	}
}
