import java.util.Arrays;
import java.util.Map;

/**
 * Semi-space copying garbage collector implementation
 *
 * @author Kuba Madej
 *
 */
public final class CollectorImpl implements Collector {
	/**
	 * Variables map
	 */
	private final Map<String, Integer> variables;

	/**
	 * Pointer up to which memory can be allocated in the lower space
	 */
	private final int lowLimit;

	/**
	 * Pointer up to which memory can be allocated in the higher space
	 */
	private final int highLimit;

	/**
	 * Pointer to the next free memory cell in current space
	 */
	private int next = HeapConstants.BASE_INDEX;

	/**
	 * Pointer up to which memory can be allocated on from-space
	 */
	private int limit;

	/**
	 * Pointer to to-space
	 */
	private int to;

	/**
	 * Allocates memory
	 *
	 * @param heap
	 *            heap to work on
	 * @return pointer to just allocated memory or 0 if out of memory
	 */
	public int allocateMemory(int heap[]) {
		if (next < limit) {
			Arrays.fill(heap, next, next + HeapConstants.OBJECT_SIZE, 0);
			final int indexToReturn = next;
			next += HeapConstants.OBJECT_SIZE;
			return indexToReturn;
		}
		return 0;
	}

	/**
	 * Constructs new instance, calculates correct values for lowLimit and
	 * highLimit.
	 *
	 * @param variables
	 *            variables map
	 */
	public CollectorImpl(final Map<String, Integer> variables,
			final int heapSize) {
		this.variables = variables;
		final int halfSize = (heapSize - 1) / 2;
		final int spaceSize = halfSize - halfSize % HeapConstants.OBJECT_SIZE;
		lowLimit = HeapConstants.BASE_INDEX + spaceSize;
		highLimit = lowLimit + spaceSize;

		next = HeapConstants.BASE_INDEX;
		limit = lowLimit;
		to = lowLimit;
	}

	/**
	 * To achieve better performance all private methods where inlined
	 *
	 * @see Collector#collect(int[])
	 */
	public final void collect(final int[] heap) {
		int scan = next = to;

		for (Map.Entry<String, Integer> entry : variables.entrySet()) {
			final int originalPointer = entry.getValue();
			final int destinationPointer = heap[originalPointer];

			final int pointerToSave;
			if (destinationPointer == 0) {
				System.arraycopy(heap, originalPointer + 1, heap, next + 1,
						HeapConstants.OBJECT_SIZE - 1);
				heap[next] = 0;
				heap[originalPointer] = next;
				next += HeapConstants.OBJECT_SIZE;
				pointerToSave = heap[originalPointer];
			} else {
				pointerToSave = destinationPointer;
			}

			variables.put(entry.getKey(), pointerToSave);
		}

		while (scan < next) {
			int pointerIndex = scan + HeapConstants.F1;
			int pointer = heap[pointerIndex];
			if (pointer != 0) {
				final int destinationPointer = heap[pointer];
				if (destinationPointer == 0) {
					System.arraycopy(heap, pointer + 1, heap, next + 1,
							HeapConstants.OBJECT_SIZE - 1);
					heap[next] = 0;
					heap[pointer] = next;
					next += HeapConstants.OBJECT_SIZE;
					heap[pointerIndex] = heap[pointer];
				} else {
					heap[pointerIndex] = destinationPointer;
				}
			}
			pointerIndex = scan + HeapConstants.F2;
			pointer = heap[pointerIndex];
			if (pointer != 0) {
				final int destinationPointer = heap[pointer];
				if (destinationPointer == 0) {
					System.arraycopy(heap, pointer + 1, heap, next + 1,
							HeapConstants.OBJECT_SIZE - 1);
					heap[next] = 0;
					heap[pointer] = next;
					next += HeapConstants.OBJECT_SIZE;
					heap[pointerIndex] = heap[pointer];
				} else {
					heap[pointerIndex] = destinationPointer;
				}
			}
			scan += HeapConstants.OBJECT_SIZE;
		}

		if (to == HeapConstants.BASE_INDEX) {
			to = lowLimit;
			limit = lowLimit;
		} else {
			to = HeapConstants.BASE_INDEX;
			limit = highLimit;
		}
	}
}
