import java.util.HashMap;
import java.util.Map;

/**
 * Class for transforming statements from their textual to objective
 * representation
 *
 * @author Kuba Madej
 *
 */
public class StatementFactory {
	/**
	 * String representing first reference field in an object
	 */
	private static final String F2_STRING = "f2";

	/**
	 * String representing second reference field in an object
	 */
	private static final String F1_STRING = "f1";

	/**
	 * Assignment operator
	 */
	private static final char ASSIGNMENT_OPERATOR = '=';

	/**
	 * Dereference operator
	 */
	private static final String DEREFERENCE_OPERATOR = ".";

	/**
	 * Variable declaration statement
	 */
	private static final String VAR_DECL_STATEMENT_STRING = "VarDecl ";

	/**
	 * Collect statement
	 */
	private static final String COLLECT_STATEMENT_STRING = "Collect";

	/**
	 * Heap analyze statement
	 */
	private static final String HEAP_ANALYZE_STATEMENT_STRING = "HeapAnalyze";

	/**
	 * Print statement
	 */
	private static final String PRINT_STATEMENT_STRING = "Print ";

	/**
	 * Empty statement
	 */
	private static final String EMPTY_STATEMENT_STRING = "";

	/**
	 * Heap to work on
	 */
	private int[] heap;

	/**
	 * Variables map
	 */
	private Map<String, Integer> variables = new HashMap<String, Integer>();

	/**
	 * Garbage collector to be used
	 */
	private CollectorImpl collector;

	/**
	 * Constructs new instance
	 *
	 * @param heap
	 *            heap to be used by statements produced
	 */
	public StatementFactory(int[] heap) {
		this.heap = heap;
		collector = new CollectorImpl(variables, heap.length);
	}

	/**
	 * Method for transforming statements from their textual to objective
	 * representation
	 *
	 * @param statementString
	 *            string representing statement to be parsed
	 * @return parsed statement
	 */
	public Statement getStatement(final String statementString) {

		if (EMPTY_STATEMENT_STRING.equals(statementString))
			return EmptyStatement.getInstance();

		if (statementString.startsWith(PRINT_STATEMENT_STRING)) {
			final String trimmedText = statementString.substring(
					PRINT_STATEMENT_STRING.length()).trim();
			return new PrintStatement(trimmedText.substring(1, trimmedText
					.length() - 1));
		}

		if (HEAP_ANALYZE_STATEMENT_STRING.equals(statementString))
			return new HeapAnalyzeStatement(heap);

		if (COLLECT_STATEMENT_STRING.equals(statementString))
			return new CollectStatement(heap, collector);

		if (statementString.startsWith(VAR_DECL_STATEMENT_STRING))
			return new VariableDeclarationStatement(heap, variables, collector,
					statementString.substring(
							VAR_DECL_STATEMENT_STRING.length()).trim());

		if (statementString.indexOf(ASSIGNMENT_OPERATOR) != -1) {
			final String[] assignment = statementString
					.split(ASSIGNMENT_OPERATOR + "");
			final String[] lValue = assignment[0].trim().split(
					"\\" + DEREFERENCE_OPERATOR);
			final String[] rValue = assignment[1].trim().split(
					"\\" + DEREFERENCE_OPERATOR);
			return new AssignmentStatement(heap, variables, lValue[0],
					createDereferenceArray(lValue), rValue[0],
					createDereferenceArray(rValue));
		}

		throw new RuntimeException("Unrecognised statement: " + statementString);
	}

	/**
	 * Creates an array containing offsets of fields referenced
	 *
	 * @param path
	 *            string representations of objects' fields
	 * @return an array containing offsets of fields referenced
	 */
	private int[] createDereferenceArray(String[] path) {
		final int[] valueDeref = new int[path.length - 1];
		for (int i = 1; i < path.length; i++) {
			if (F1_STRING.equals(path[i]))
				valueDeref[i - 1] = HeapConstants.F1;
			else if (F2_STRING.equals(path[i]))
				valueDeref[i - 1] = HeapConstants.F2;
			else
				valueDeref[i - 1] = HeapConstants.DATA;
		}
		return valueDeref;
	}
}
